import importlib
import pkgutil

discovered_plugins = {
    name: importlib.import_module(name)
    for finder, name, ispkg
    in pkgutil.iter_modules()
    if name.startswith('gitflow_') and name.endswith('_linter') and name != 'gitflow_linter'
}


def validate_plugin(plugin_module):
    visitors = plugin_module.visitors(gitflow={})
    from gitflow_linter.visitor import BaseVisitor
    invalid = [visitor
               for visitor in plugin_module.visitors(gitflow={})
               if not isinstance(visitor, BaseVisitor) or not visitor.rule]
    if not visitors:
        raise ValueError('Plugin is invalid because it has no visitors')
    if len(invalid) > 0:
        raise ValueError('Plugin is invalid because it has visitors that evaluate no rule or do not extend '
                         'BaseVisitor: {} '
                         .format(', '.join([str(type(v)) for v in invalid])))


def is_plugin_valid(plugin_module) -> bool:
    try:
        validate_plugin(plugin_module=plugin_module)
    except BaseException:
        return False
    else:
        return True



import logging
import os
import sys
from os import linesep
from gitflow_linter.report import Report, Section

FORMAT = '%(message)s'
logging.basicConfig(format=FORMAT, level=logging.INFO)
log = logging.getLogger('main')

_handler = logging.StreamHandler(sys.stdout)
stdout_log = logging.getLogger('results')
stdout_log.addHandler(_handler)


def _console_output(report: Report):
    def _section_icon(s: Section) -> str:
        if s.contains_issues and not section.contains_errors and not section.contains_warns:
            return 'ℹ️'
        elif s.contains_errors:
            return '❌'
        elif s.contains_warns:
            return '⚠️'
        else:
            return '✅'

    title = 'Report for git repository: {}'.format(report.working_dir)
    log.info('=' * len(title))
    log.info(title)
    log.info('=' * len(title))
    counts_format = '\t{}: {} open branch(es)'
    log.info('Statistics:' + os.linesep + os.linesep.join([counts_format.format(branch, count) for branch, count in report.stats['counts'].items()]))
    log.info('=' * len(title))
    log.info(linesep + 'Results:')
    for section in report.sections:
        log.info(linesep + _section_icon(section) + '\t' + section.rule)
        log.info('\t' + section.title)
        if section.issues:
            log.info('\tIssues detected:')
            for issue in section.issues:
                log.log(issue.level.to_log_level, '\t\t- ' + issue.description)


def _json_output(report: Report):
    import json
    stdout_log.info(json.dumps({
        'repository': report.working_dir,
        'statistics': report.stats,
        'sections': [
            {
                'rule': section.rule,
                'title': section.title,
                'issues': [
                    {
                        'level': issue.level,
                        'description': issue.description,
                    } for issue in section.issues
                ]
            } for section in report.sections
        ]
    }, indent=2))


outputs = {
    'console': _console_output,
    'json': _json_output,
}


def create_output(out_type) -> callable:
    return outputs.get(out_type, _console_output)



import sys
import os

# GitPython, click, pyyaml
import click
from git import Repo
import yaml

from gitflow_linter import output
from gitflow_linter.rules import RulesContainer, Gitflow

DEFAULT_LINTER_OPTIONS = 'gitflow_linter.yaml'
__version__ = '0.0.4'


def _validate_settings(value, working_dir):
    if value:
        return value

    potential_settings = os.path.join(working_dir, DEFAULT_LINTER_OPTIONS)
    if not os.path.exists(potential_settings):
        raise click.BadParameter("Working git directory {} does not contain {} file. ".format(working_dir,
                                                                                              DEFAULT_LINTER_OPTIONS) +
                                 "Please provide path to the settings by using --settings option")
    return open(potential_settings, 'r')


@click.command()
@click.argument('git_directory',
                type=click.Path(exists=True, file_okay=False, dir_okay=True, readable=True, resolve_path=True,
                                allow_dash=False, path_type=None))
@click.option('-s', '--settings', type=click.File(mode='r', encoding=None, errors='strict', lazy=None, atomic=False))
@click.option('-o', '--output', 'out',
              type=click.Choice(output.outputs.keys(), case_sensitive=False), default=next(iter(output.outputs.keys())))
@click.option('-p', '--fetch-prune', 'fetch', is_flag=True, default=False, help="Linter will refresh the repo before "
                                                                                "checking")
@click.option('-d', '--allow-dirty', is_flag=True, default=False, help="Linter will ignore the fact that the given "
                                                                       "repo is considered dirty")
def main(git_directory, settings, out, fetch, allow_dirty):
    """Evaluate given repository and check if gitflow is respected"""
    from gitflow_linter.report import Report, Section, Issue
    from gitflow_linter.visitor import StatsRepositoryVisitor
    from gitflow_linter.repository import Repository

    try:
        settings = _validate_settings(settings, working_dir=git_directory)
        gitflow, rules = parse_yaml(settings)
        repo = Repository(Repo(git_directory), gitflow=gitflow, should_fetch=fetch, allow_dirty=allow_dirty)
        report = Report(working_dir=git_directory, stats=repo.apply(StatsRepositoryVisitor(gitflow=gitflow)), sections=[])

        visitors = __get_all_visitors(gitflow=gitflow, rules=rules)
        for visitor in visitors.values():
            try:
                kwargs = rules.args_for(visitor.rule)
                section: Section = repo.apply(visitor, **kwargs if kwargs else {})
                if section is not None:
                    report.append(section)
                else:
                    output.log.warning('⚠️ Rule {} checked but result was not returned'.format(visitor.rule))
            except BaseException as err:
                error_section = Section(rule=visitor.rule, title='ERROR!')
                error_section.append(Issue.error('💀 Cannot be checked because of error: {err}'.format(err=err)))
                report.append(error_section)
            finally:
                rules.consume(visitor.rule)

        if rules.rules:
            output.log.warning('Some of rules cannot be validated because corresponding validators could not be found: '
                               + ', '.join(rules.rules))
        output.create_output(out)(report)
        return sys.exit(1 if report.contains_errors(are_warnings_errors=False) else 0)
    except BaseException as err:
        output.log.error(err)
        return sys.exit(1)


def parse_yaml(settings):
    yaml_settings = yaml.load(settings, Loader=yaml.SafeLoader)
    gitflow = Gitflow(settings=yaml_settings)
    rules = RulesContainer(rules=yaml_settings)
    return gitflow, rules


def __get_all_visitors(gitflow: Gitflow, rules: RulesContainer) -> dict:
    from gitflow_linter import visitor
    from gitflow_linter import plugins
    visitors = [visitor for visitor in visitor.visitors(gitflow=gitflow) if visitor.rule in rules.rules]
    plugin_visitors = [plugin.visitors(gitflow=gitflow)
                       for plugin in plugins.discovered_plugins.values()
                       if plugins.is_plugin_valid(plugin_module=plugin)]
    flatten = lambda t: [item for sublist in t for item in sublist]
    all_visitors = visitors + [plugin_visitor
                               for plugin_visitor in flatten(plugin_visitors)
                               if plugin_visitor.rule in rules.rules]
    return {
        v.rule: v
        for v in all_visitors
    }


@click.command()
def available_plugins():
    from gitflow_linter import plugins
    available_plugins = plugins.discovered_plugins.keys()
    output.stdout_log.info('Available gitflow-linter plugins:')
    if not available_plugins:
        output.stdout_log.info('No plugins found.')
    for plugin in available_plugins:
        try:
            plugins.validate_plugin(plugin_module=plugins.discovered_plugins[plugin])
            plugin_visitors = plugins.discovered_plugins[plugin].visitors(gitflow={})
            log_fmt = '- {} handles following rules: ' + os.linesep + '\t* {}'
            output.stdout_log.info(log_fmt.format(plugin, '\t* '.join([v.rule for v in plugin_visitors])))
        except BaseException as err:
            output.stdout_log.error('❌ {} cannot be used because of error: {}'.format(plugin, err))
    return sys.exit(0)


if __name__ == '__main__':
    main()



import logging
from typing import List
from enum import Enum, unique


@unique
class Level(str, Enum):
    INFO = 'info'
    WARNING = 'warning'
    ERROR = 'error'

    @property
    def to_log_level(self):
        return {
            Level.INFO: logging.INFO,
            Level.WARNING: logging.WARNING,
            Level.ERROR: logging.ERROR
        }.get(self, logging.DEBUG)


class Issue:

    @classmethod
    def info(cls, description: str):
        """
        Creates an ``Issue`` with INFO severity
        """
        return cls(Level.INFO, description)

    @classmethod
    def warning(cls, description: str):
        """
        Creates an ``Issue`` with WARNING severity
        """
        return cls(Level.WARNING, description)

    @classmethod
    def error(cls, description: str):
        """
        Creates an ``Issue`` with ERROR severity
        """
        return cls(Level.ERROR, description)

    def __init__(self, level: Level, description: str):
        """
        :param level: Describes severity of the Issue
        :param description: Explanation of what is wrong
        """
        self.level = level
        self.description = description

    def __repr__(self):
        return "Issue(level={level}, description='{desc}')".format(level=self.level, desc=self.description)


class Section:
    """
    Represents repository verification done for a single rule.
    Results are represented by list of :class:`Issues <Issue>`.
    """

    def __init__(self, rule: str, title: str, issues=None):
        if issues is None:
            issues = []
        self.rule = rule
        self.title = title
        self.issues = issues

    def append(self, issue: Issue):
        """
        Adds new issue detected

        :param issue: New issue detected
        :return:
        """
        self.issues.append(issue)

    def extend(self, issues: list):
        self.issues.extend(issues)

    @property
    def contains_issues(self) -> bool:
        return len(self.issues) > 0

    @property
    def contains_errors(self) -> bool:
        return len([issue for issue in self.issues if issue.level is Level.ERROR]) > 0

    @property
    def contains_warns(self) -> bool:
        return len([issue for issue in self.issues if issue.level is Level.WARNING]) > 0

    def __repr__(self):
        return "Section(rule={rule}, title='{title}')".format(rule=self.rule, title=self.title)


class Report:
    def __init__(self, working_dir: str, stats: dict, sections: list):
        self.working_dir = working_dir
        self.stats = stats
        self.sections = sections if sections else []

    def append(self, section: Section):
        self.sections.append(section)

    def contains_errors(self, are_warnings_errors) -> bool:
        errors = [section for section in self.sections if
                  section.contains_errors or (are_warnings_errors and section.contains_warns)]
        return len(errors) > 0



class AttributeDict(dict):
    __slots__ = ()
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__

    def __init__(self, dictionary: dict = {}):
        super().__init__()
        self.update(dictionary)


_defaults = AttributeDict({
        "master": "master",
        "develop": "develop",
        "features": "feature",
        "fixes": "bugfix",
        "releases": "release",
        "hotfixes": "hotfix",
        "others": ["spike"],
    })


class Gitflow(AttributeDict):
    """
    Contains all settings related to branches from YAML file.
    It extends :class:`AttributeDict <AttributeDict>`, so settings may be accessed like properties: ``gitflow.features``

    It will contain custom settings if you add them in YAML file as a child of ``branches`` node.
    """

    def __init__(self, settings: dict):
        defaults = _defaults.copy()
        if settings and settings.get('branches', None):
            defaults.update(settings['branches'])
        super().__init__(dictionary=defaults)


class RulesContainer:

    def __init__(self, rules: dict):
        if not rules or not rules.get('rules', None):
            raise KeyError('Yaml file does not contain rules')
        self._rules = rules['rules']

    @property
    def rules(self):
        return self._rules.keys()

    def args_for(self, rule) -> dict:
        return self._rules.get(rule, {})

    def consume(self, rule: str):
        del self._rules[rule]



from abc import ABC, abstractmethod
from datetime import datetime, timedelta
import os
from typing import List
from functools import wraps

from git import Head
from git.util import IterableList

from gitflow_linter import Gitflow
from gitflow_linter.report import Section, Issue, Level
from gitflow_linter.repository import Repository, RepositoryVisitor


def arguments_checker(keywords):
    def wrap(f):
        @wraps(f)
        def new_function(*args, **kw):
            missing = [keyword for keyword in keywords if keyword not in kw.keys()]
            if len(missing) > 0:
                raise ValueError(
                    "Following arguments are missing: {}".format(', '.join(keywords)))
            return f(*args, **kw)

        return new_function

    return wrap


class BaseVisitor(RepositoryVisitor, ABC):
    """
    Abstract class describing how gitflow-linter works. A visitor must provide a rule that it is supposed to verify.
    The linter will let the visitor visit a repository only if user wants to check the repository against the rule.
    Plugins can override default visitors by returning the same rule as a visitor they wish override.
    """

    @property
    @abstractmethod
    def rule(self) -> str:
        """
        :return: Rule from YAML file that is checked by the visitor
        """
        pass

    def __init__(self, gitflow: Gitflow):
        self.gitflow = gitflow

    @abstractmethod
    def visit(self, repo: Repository, *args, **kwargs) -> Section:
        """
        Verifies the :class:`repository <gitflow_linter.repository.Repository>` - checks if ``self.rule`` is respected

        :param repo: Tiny wrapper for GitPython's repository
        :param args:
        :param kwargs: arguments from YAML file
        :return: :class:`Section <gitflow_linter.report.Section>` with results
        """
        pass


class StatsRepositoryVisitor(RepositoryVisitor):

    def __init__(self, gitflow: Gitflow):
        self.gitflow = gitflow

    def visit(self, repo: Repository, **kwargs):
        def names(branches: IterableList):
            return [b.name for b in branches]

        return {
            "references": {
                "master": names(branches=repo.branches(folder=self.gitflow.master)),
                "develop": names(branches=repo.branches(folder=self.gitflow.develop)),
                "features": names(branches=repo.branches(folder=self.gitflow.features)),
                "fixes": names(branches=repo.branches(folder=self.gitflow.fixes)),
                "releases": names(branches=repo.branches(folder=self.gitflow.releases)),
                "hotfixes": names(branches=repo.branches(folder=self.gitflow.hotfixes)),
            },
            "counts": {
                "master": len(repo.branches(folder=self.gitflow.master)),
                "develop": len(repo.branches(folder=self.gitflow.develop)),
                "features": len(repo.branches(folder=self.gitflow.features)),
                "fixes": len(repo.branches(folder=self.gitflow.fixes)),
                "releases": len(repo.branches(folder=self.gitflow.releases)),
                "hotfixes": len(repo.branches(folder=self.gitflow.hotfixes)),
            }
        }


class SingleBranchesVisitor(BaseVisitor):
    __doc__ = """gitflow strongly relies on the fact that there is (1) only one branch for keeping the release history 
    and (2) only one integration branch """

    @property
    def rule(self) -> str:
        return 'single_master_and_develop'

    def visit(self, repo: Repository, **kwargs) -> Section:
        section = Section(rule=self.rule, title='Checked if repo contains single release history branch and single '
                                                'integration branch')
        # TODO add smarter checking
        if len(repo.branches(folder=self.gitflow.master)) > 1:
            section.append(Issue.error('Repository contains more than one master branch'))
        if len(repo.branches(folder=self.gitflow.develop)) > 1:
            section.append(Issue.error('Repository contains more than one develop branch'))

        return section


class OldDevelopmentBranchesVisitor(BaseVisitor):
    __doc__ = """having old feature or bugfix branches may create a mess in the repository
    
    use ``max_days_features`` option to configure what 'old' means for you"""

    @property
    def rule(self) -> str:
        return 'no_old_development_branches'

    @arguments_checker(['max_days_features'])
    def visit(self, repo: Repository, **kwargs) -> Section:
        section = Section(rule=self.rule, title='Checked if repo contains abandoned feature branches')
        deadline = datetime.now() - timedelta(days=kwargs['max_days_features'])
        merged_branches = repo.raw_query(lambda git: git.branch('-r', '--merged', self.gitflow.develop))

        def _check_for_issues(branches: IterableList, name: str):
            for branch in branches:
                if deadline > branch.commit.authored_datetime.replace(tzinfo=None) \
                        and branch.name not in merged_branches:
                    section.append(Issue.error(
                        '{} {} has not been touched since {}'.format(name, branch.name,
                                                                     branch.commit.authored_datetime)))

        _check_for_issues(branches=repo.branches(folder=self.gitflow.features), name='Feature')
        _check_for_issues(branches=repo.branches(folder=self.gitflow.fixes), name='Fix')

        return section


class NotScopedBranchesVisitor(BaseVisitor):
    __doc__ = """having branches that are out of configured folders (eg. created out of feature/, bugfix/) may be an 
    indicator that you do something wrong and create a mess in the repo"""

    @property
    def rule(self) -> str:
        return 'no_orphan_branches'

    def visit(self, repo: Repository, **kwargs) -> Section:
        section = Section(rule=self.rule, title='Checked if repo contains orphan branches (created out of expected '
                                                'folders)')
        expected_prefix_template = '{remote}/{branch}'
        expected_prefixes = [
                                expected_prefix_template.format(remote=repo.remote.name, branch='HEAD'),
                                expected_prefix_template.format(remote=repo.remote.name,
                                                                branch=self.gitflow.master),
                                expected_prefix_template.format(remote=repo.remote.name,
                                                                branch=self.gitflow.develop),
                                expected_prefix_template.format(remote=repo.remote.name,
                                                                branch=self.gitflow.features),
                                expected_prefix_template.format(remote=repo.remote.name,
                                                                branch=self.gitflow.fixes),
                                expected_prefix_template.format(remote=repo.remote.name,
                                                                branch=self.gitflow.hotfixes),
                                expected_prefix_template.format(remote=repo.remote.name,
                                                                branch=self.gitflow.releases),
                            ] + [expected_prefix_template.format(remote=repo.remote.name, branch=branch.strip())
                                 for branch in self.gitflow.others]

        def has_expected_prefix(branch: Head) -> bool:
            for prefix in expected_prefixes:
                if branch.name.startswith(prefix):
                    return True
            return False

        orphan_branches = [branch for branch in repo.branches() if not has_expected_prefix(branch=branch)]
        for branch in orphan_branches:
            section.append(Issue.error('{branch} looks like created out of expected scopes'.format(branch=branch.name)))

        return section


class MainCommitsAreTaggedVisitor(BaseVisitor):
    __doc__ = """if your master branch contains commits that are not tagged, it probably means that you don't use 
    master as your releases history keeper"""

    @property
    def rule(self) -> str:
        return 'master_must_have_tags'

    def visit(self, repo: Repository, **kwargs) -> Section:
        section = Section(rule=self.rule, title='Checked if main repo branch has tagged commits')
        main_branch = '/'.join([repo.remote.name, self.gitflow.master])
        main_commits = repo.raw_query(
            lambda git: git.log(main_branch, '--merges', '--format=format:%H', '--first-parent'),
            predicate=lambda sha: sha)
        tags = repo.repo.tags
        tags_sha = [tag.commit.hexsha for tag in tags]
        tags_not_on_main_branch = [sha for sha in tags_sha if sha not in main_commits]
        main_commits_not_tagged = [commit for commit in main_commits if commit not in tags_sha]

        for main_commit_not_tagged in main_commits_not_tagged:
            section.append(Issue.error('{commit} commit in main branch is not tagged'
                                       .format(commit=main_commit_not_tagged[:8])))

        for tag_not_on_main in tags_not_on_main_branch:
            section.append(Issue.warning('{commit} commit contains a tag but is not a part of the master branch'
                                         .format(commit=tag_not_on_main[:8])))

        return section


class NoDirectCommitsToProtectedBranches(BaseVisitor):
    """
    the purposes behind develop and master are different but there is an assumption that at least those two are protected

    the rule is here to check if it is really the case and both branches does not contain direct commits (commits
    that were pushed directly)
    """

    @property
    def rule(self) -> str:
        return 'no_direct_commits_to_protected_branches'

    def visit(self, repo: Repository, *args, **kwargs) -> Section:
        section = Section(rule=self.rule, title='Checked if {} and {} contain only merges without direct commits'
                          .format(self.gitflow.develop, self.gitflow.master))
        initial_commit = next(iter(repo.raw_query(lambda git: git.log('--format=format:%s', '--reverse'))))

        def _get_direct_commits(branch: str) -> List[str]:
            merges = repo.raw_query(
                lambda git: git.log(branch, '--merges', '--format=format:%H', '--first-parent'),
                predicate=lambda sha: sha)
            potential_fast_forwards = repo.raw_query(
                lambda git: git.reflog('show', branch, '--format=format:%H'),
                predicate=lambda sha: sha)
            all_commits = repo.raw_query(
                lambda git: git.log(branch, '--format=format:%H', '--first-parent'),
                predicate=lambda sha: sha)
            return [sha for sha in all_commits if sha not in merges and sha not in potential_fast_forwards]

        def _get_issues(direct_commits: List[str], branch: str) -> List[Issue]:
            issuers = [repo.commit(sha, branch) for sha in direct_commits]
            issue_msg_fmt = 'Branch {} contains commit "{}" that was pushed directly rather than merged'
            return [
                Issue.error(issue_msg_fmt.format(branch, ' '.join([str(commit.hexsha)[:8], commit.message.strip()])))
                for commit in issuers if commit and commit.message.lower().strip() != initial_commit.lower().strip()
            ]

        develop_results = _get_direct_commits(self.gitflow.develop)
        master_results = _get_direct_commits(self.gitflow.master)
        section.extend(_get_issues(direct_commits=develop_results, branch=self.gitflow.develop))
        section.extend(_get_issues(direct_commits=master_results, branch=self.gitflow.master))
        return section


class VersionNamesConventionVisitor(BaseVisitor):
    __doc__ = """checks if release branches and tags follow version naming convention
    
    the convention must be specified in ``version_regex`` argument as a regular expression string"""

    @property
    def rule(self) -> str:
        return 'version_names_follow_convention'

    @arguments_checker(['version_regex'])
    def visit(self, repo: Repository, *args, **kwargs) -> Section:
        import re
        section = Section(rule=self.rule, title='Checked if version names follow given convention')
        releases = [branch.name for branch in repo.branches(self.gitflow.releases)]
        tags = [tag.name for tag in repo.repo.tags]
        version_reg = kwargs['version_regex']

        def _validate_version(v: str) -> bool:
            return re.search(version_reg, v) is not None

        release_issues = [release for release in releases if not _validate_version(release.split('/')[-1])]
        tags_issues = [tag for tag in tags if not _validate_version(tag)]

        section.extend(
            [Issue.error('Release {branch} does not follow name convention'.format(branch=release)) for release in
             release_issues])
        section.extend(
            [Issue.error('Tag {tag} does not follow name convention'.format(tag=tag)) for tag in tags_issues])

        return section


class DevBranchNamesFollowConvention(BaseVisitor):
    """
    sometimes you may wish to have feature and bugfix branch names containing eg. ticket numbers

    the given convention is checked by providing ``name_regex`` as an argument

    if you want to provide different conventions for features and bugfixes, use ``feature_name_regex`` and ``bugfix_name_regex`` respectively
    """

    @property
    def rule(self) -> str:
        return 'dev_branch_names_follow_convention'

    def visit(self, repo: Repository, *args, **kwargs) -> Section:
        import re
        features_regex = kwargs.get('name_regex', None) if not kwargs.get('feature_name_regex', None) else kwargs.get(
            'feature_name_regex')
        bugfixes_regex = kwargs.get('name_regex', None) if not kwargs.get('bugfix_name_regex', None) else kwargs.get(
            'bugfix_name_regex')
        if not features_regex or not bugfixes_regex:
            raise Exception('Configuration of the rule is invalid: desired convention is not provided')

        feature_prefix = '/'.join([repo.remote.name, self.gitflow.features, ''])
        bugfix_prefix = '/'.join([repo.remote.name, self.gitflow.fixes, ''])

        def _is_valid(branch: str, regex: str) -> bool:
            return re.search(regex, branch) is not None

        feature_issuers = [feature.name for feature in repo.branches(self.gitflow.features) if
                           not _is_valid(branch=feature.name.replace(feature_prefix, ''), regex=features_regex)]
        bugfix_issuers = [bugfix.name for bugfix in repo.branches(self.gitflow.fixes) if
                          not _is_valid(branch=bugfix.name.replace(bugfix_prefix, ''), regex=bugfixes_regex)]

        issue_msg_fmt = '{branch} branch does not follow given convention'
        issues = [Issue.error(issue_msg_fmt.format(branch=branch)) for branch in (feature_issuers + bugfix_issuers)]

        return Section(rule=self.rule, title='Checked if feature and bugfix names follow convention', issues=issues)


class DeadReleasesVisitor(BaseVisitor):
    __doc__ = """release branches that are not closed may create a mess in the repository and breaks the master/main 
    branch - releases must be closed as soon as they are deployed to production environment (or just before, 
    depending on your case)
    
    since hotfixes are in fact releases started from master instead of develop, the rule will be checked against them as well
    
    configure how long releases are supposed to be maintained by using ``deadline_to_close_release`` (number of days)"""

    @property
    def rule(self) -> str:
        return 'no_dead_releases'

    @arguments_checker(['deadline_to_close_release'])
    def visit(self, repo: Repository, *args, **kwargs) -> Section:
        section = Section(rule=self.rule, title='Checked if repo contains abandoned and not removed releases')
        deadline = datetime.now() - timedelta(days=kwargs['deadline_to_close_release'])
        main_branch = '/'.join([repo.remote.name, self.gitflow.master])
        release_branch = '/'.join([repo.remote.name, self.gitflow.releases, ''])
        hotfix_branch = '/'.format([repo.remote.name, self.gitflow.hotfixes, ''])

        potential_dead_releases = repo.raw_query(lambda git: git.branch('-r', '--no-merged', main_branch),
                                                 predicate=lambda release: release.strip().startswith(
                                                     release_branch) or release.strip().startswith(
                                                     hotfix_branch),
                                                 map_line=lambda release: repo.branch(release))
        dead_releases = [dead_release for dead_release in potential_dead_releases if
                         deadline > dead_release.commit.authored_datetime.replace(tzinfo=None)]

        section.extend([Issue.error(
            '{release} seems abandoned - it has never been merged into the master branch'.format(release=r.name)) for r
            in
            dead_releases])

        return section


class DependantFeaturesVisitor(BaseVisitor):
    __doc__ = """creating feature/bugfix branches one from another or merging them together before merging to develop 
    may result in ugly issues during code review and merge mistakes 
    
    creating such a feature/merge is sometimes inevitable, you must configure the limit of such branches by using 
    ``max_dependant_branches`` option """

    @property
    def rule(self) -> str:
        return 'no_dependant_features'

    @arguments_checker(['max_dependant_branches'])
    def visit(self, repo: Repository, *args, **kwargs) -> Section:
        section = Section(rule=self.rule, title='Checked if repo contains dependant feature branches')
        dev_branch = '/'.join([repo.remote.name, self.gitflow.develop])
        max_dependant_branches = int(kwargs['max_dependant_branches'])
        merged_branches = repo.raw_query(lambda git: git.branch('-r', '--merged', self.gitflow.develop))
        not_merged = [repo.branch(b.name) for b in repo.branches(self.gitflow.features) if
                      b.name not in merged_branches]
        branch_issue_format = '{} seems to depend on other feature branches. It contains following merges: ' + os.linesep + '{}'

        for feature in not_merged:
            name = feature.name
            merge_commits_query = repo.raw_query(lambda git: git.log('..'.join([dev_branch, name]), '--merges',
                                                                     '--first-parent', '--format=format:%H'))
            merge_commits_sha = [commit_sha.strip() for commit_sha in merge_commits_query]
            merge_commits_in_feature = [commit for commit in repo.repo.iter_commits(name, max_count=200) if
                                        commit.hexsha in merge_commits_sha]
            branch_issues = [commit for commit in merge_commits_in_feature if
                             self.gitflow.develop not in commit.message]

            if branch_issues:
                is_limit_exceeded = len(branch_issues) > max_dependant_branches
                issue_level = Level.ERROR if is_limit_exceeded else Level.WARNING
                issues_titles = [next(iter(commit.message.split(os.linesep)), None) for commit in branch_issues]
                issue_desc = branch_issue_format.format(name, os.linesep.join(issues_titles))
                section.append(Issue(level=issue_level, description=issue_desc))

        return section


def visitors(gitflow: Gitflow) -> List[BaseVisitor]:
    return [
        SingleBranchesVisitor(gitflow=gitflow),
        OldDevelopmentBranchesVisitor(gitflow=gitflow),
        NotScopedBranchesVisitor(gitflow=gitflow),
        MainCommitsAreTaggedVisitor(gitflow=gitflow),
        NoDirectCommitsToProtectedBranches(gitflow=gitflow),
        VersionNamesConventionVisitor(gitflow=gitflow),
        DevBranchNamesFollowConvention(gitflow=gitflow),
        DeadReleasesVisitor(gitflow=gitflow),
        DependantFeaturesVisitor(gitflow=gitflow),
    ]



import os
from abc import ABC, abstractmethod
from typing import Optional

from git import Repo, Remote, RemoteReference, Commit, Head
from git.util import IterableList

from gitflow_linter import Gitflow


class Repository:
    def __init__(self, repo: Repo, gitflow: Gitflow, should_fetch=False, allow_dirty=False):
        self.repo = repo
        self.gitflow = gitflow
        self.assert_repo(allow_dirty)
        if should_fetch:
            self.remote.fetch('--prune')

    def assert_repo(self, allow_dirty: bool):
        if self.repo.bare:
            raise Exception('Given directory {} does not contain valid GIT repository.'.format(self.repo.working_dir))
        if not self.branch(self.gitflow.develop):
            raise Exception('Given repository {} does not contain expected {} branch'.format(self.repo.working_dir, self.gitflow.develop))
        if not self.branch(self.gitflow.master):
            raise Exception('Given repository {} does not contain expected {} branch'.format(self.repo.working_dir, self.gitflow.master))
        if self.repo.is_dirty(untracked_files=False) and not allow_dirty:
            raise Exception('Given repository {} is dirty.'.format(self.repo.working_dir))
        if len(self.repo.remotes) > 1:
            raise Exception(
                'Repo contains more than one remote: [{}]'.format(', '.join([r.name for r in self.repo.remotes])))

    @property
    def remote(self) -> Remote:
        return self.repo.remotes[0]

    def branches(self, folder: str = None) -> IterableList:
        path = None if folder is None else '{}/{}'.format(self.remote.name, folder)
        return self.remote.refs if path is None else [r for r in self.remote.refs if path in r.name]

    def branch(self, name: str = None) -> RemoteReference:
        path = name if '/' in name else '{}/{}'.format(self.remote.name, name)
        return next(iter([r for r in self.remote.refs if r.name.startswith(path)]), None)

    @property
    def master(self) -> Head:
        return self.repo.heads[self.gitflow.master]

    @property
    def develop(self) -> Head:
        return self.repo.heads[self.gitflow.develop]

    def unique_commits_for_branch(self, branch: RemoteReference, force_including_head=True) -> set[Commit]:
        """
        Returns set of unique commits for the given branch. Only commits that appear specifically on the branch will be
        returned. If a commit is included in the given branch and any other branch, it won't be taken into
        consideration.

        :param branch: the commits specific to the branch will be returned
        :param force_including_head: the flag will force including head commit of the branch. If a second branch has been started from branch passed as param, the flag set to True will ensure, that at least the one head commit will be returned, otherwise second/child branch will "consume" all commits and none will be considered as unique for the given branch.
        :return: set of unique commits for branch passed as the

        """
        other_heads = [head for head in self.branches() if not head.name == branch.name]
        all_commits = list(self.repo.iter_commits(branch.name, max_count=100))
        commits = set(all_commits)

        def _remove_commits_exist_in(other_branch: RemoteReference):
            other_branch_commits = self.repo.iter_commits(other_branch.name, max_count=100)
            for c in other_branch_commits:
                element = next((commit for commit in commits if commit.hexsha == c.hexsha), None)
                if element and (not force_including_head or element.hexsha != branch.commit.hexsha):
                    commits.remove(element)

        for other_branch in other_heads:
            _remove_commits_exist_in(other_branch=other_branch)
            if not commits:
                break

        return commits

    def raw_query(self, query: callable, predicate: callable = None, map_line: callable = None):
        """
        Let you run raw queries on GitPython's git object

        :param query: callable where you can run raw query,
            eg. ``lambda git: git.log('master')``
        :param predicate: optional callable where you can decide if given line should be included,
            eg. ``lambda commit: commit.startwith('Merged')``. All lines are included if predicate is not given.
        :param map_line: optional callable where you can map line to other object,
            eg. when query returns list of name of branches, you can map them to branch objects: ``lambda line: repo.branch(line)``
        :return: list of lines returned by query that matches optional predicate,
            eg. ``["sha-of-commit1", "sha-of-commit2", ...]``
        """
        return [line.strip() if not map_line else map_line(line.strip())
                for line in query(self.repo.git).split(os.linesep)
                if predicate is None or predicate(line)]

    def commit(self, sha: str, branch_name: str) -> Optional[Commit]:
        branch = self.branch(branch_name)
        if branch:
            return next(iter([commit for commit in list(self.repo.iter_commits(branch.name, max_count=1000)) if commit.hexsha == sha]), None)
        return None

    def apply(self, visitor, *args, **kwargs):
        return visitor.visit(self, *args, **kwargs)


class RepositoryVisitor(ABC):

    @abstractmethod
    def visit(self, repo: Repository, *args, **kwargs):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from setuptools import setup, find_packages
import os


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='gitflow_linter',
    version='0.0.4',
    description='Checks if GitFlow is respected in a given repository, considering provided rules',
    long_description=read("README.md"),
    long_description_content_type='text/markdown',
    url='https://github.com/fighterpoul/gitflow_linter',
    author='Poul Fighter',
    author_email='fighter.poul@gmail.com',
    packages=find_packages(exclude=["tests"]),
    entry_points={
        'console_scripts': [
            'gitflow-linter = gitflow_linter:main',
            'gitflow-linter-plugins = gitflow_linter:available_plugins',
        ],
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
    ],
    include_package_data=True,
    python_requires='>3.8',
    license="MIT",
    install_requires=[
        'PyYAML>=5.4.1',
        'GitPython>=3.1.17',
        'click>=7',
    ]
)
